Package: pipeComp
Type: Package
Title: pipeComp pipeline benchmarking framework
Version: 1.22.0
Depends: R (>= 4.1)
Authors@R: c(
	person("Pierre-Luc", "Germain", email="pierre-luc.germain@hest.ethz.ch", role=c("cre","aut"), comment = c(ORCID = "0000-0003-3418-4218")), 
	person("Anthony", "Sonrel", email="anthony.sonrel@uzh.ch", role="aut", comment = c(ORCID = "0000-0002-2414-715X")), 
	person("Mark D.", "Robinson", email="mark.robinson@imls.uzh.ch", role=c("aut","fnd"), comment = c(ORCID = "0000-0002-3048-5518")) )
Description: A simple framework to facilitate the comparison of pipelines involving various steps and parameters. 
	The `pipelineDefinition` class represents pipelines as, minimally, a set of functions consecutively executed on 
	the output of the previous one, and optionally accompanied by step-wise evaluation and aggregation functions. 
	Given such an object, a set of alternative parameters/methods, and benchmark datasets, the `runPipeline` function 
	then proceeds through all combinations arguments, avoiding recomputing the same step twice and compiling evaluations 
	on the fly to avoid storing potentially large intermediate data.
Imports: BiocParallel, S4Vectors, ComplexHeatmap, SingleCellExperiment,
        SummarizedExperiment, Seurat, matrixStats, Matrix, cluster,
        aricode, methods, utils, dplyr, grid, scales, scran,
        viridisLite, clue, randomcoloR, ggplot2, cowplot,
        intrinsicDimension, scater, knitr, reshape2, stats, Rtsne,
        uwot, circlize, RColorBrewer
Suggests: BiocStyle, rmarkdown
License: GPL
Encoding: UTF-8
RoxygenNote: 7.1.1
VignetteBuilder: knitr
URL: https://doi.org/10.1186/s13059-020-02136-7
BugReports: https://github.com/plger/pipeComp
biocViews: GeneExpression, Transcriptomics, Clustering,
        DataRepresentation
git_url: https://git.bioconductor.org/packages/pipeComp
git_branch: RELEASE_3_23
git_last_commit: 843ce01
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 23:53:52 UTC; biocbuild
Author: Pierre-Luc Germain [cre, aut] (ORCID:
    <https://orcid.org/0000-0003-3418-4218>),
  Anthony Sonrel [aut] (ORCID: <https://orcid.org/0000-0002-2414-715X>),
  Mark D. Robinson [aut, fnd] (ORCID:
    <https://orcid.org/0000-0002-3048-5518>)
Maintainer: Pierre-Luc Germain <pierre-luc.germain@hest.ethz.ch>
