% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{exampleResults}
\alias{exampleResults}
\title{Example pipeline results}
\value{
a list.
}
\description{
Example benchmarking results from a scRNAseq pipeline (see vignette 
`pipeComp_scRNA`).
}
