% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{stableG}
\alias{stableG}
\title{Lists of stable genes}
\value{
a list.
}
\description{
Genes were simply obtained by querying the respective GO terms
}
