% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{exampleDEAresults}
\alias{exampleDEAresults}
\title{Example results from the DEA pipeline}
\value{
a list.
}
\description{
Example benchmarking results from a DEA pipeline (see vignette 
`pipeComp_dea`).
}
